/*
CDBeQuiet! - Slows down your optical drive under DOS
Copyright (C) 2017  Christoph Boelitz

E-mail: c.boelitz@web.de
Project page: http://cdbq.dosforum.de/

CDBeQuiet! is fully free of charge and may be
distributed. Its source code may be modified as long
as you leave my contact information intact. Using it
in any commercial manner is prohibited.

Let me know, if you have suggestions or see something that can
be done in a better way, so I might learn from it. Thanks!

*/

//Type definitions
typedef unsigned short PORT;
typedef unsigned char BYTE;
typedef unsigned short WORD;

typedef struct drive_info
{
  char firmware_revision[9];  //8 chars + terminating 0 char
  char model_number[41];    //40 chars + terminating 0 char
  WORD general_config;
} drive_info;

//Defines

//Global errors//IDE errors
#define IDE_ERR_SUCCESS                100
#define IDE_ERR_TIME                   101
#define IDE_ERR_TIMEOUT                102 //Used internally, no screen message
#define IDE_ERR_WAITBUSY               103
#define IDE_ERR_WAITDRQ                104
#define IDE_ERR_WAITDRDY               105
#define IDE_ERR_IDDRIVE_NOT_ABORTED    106
#define IDE_ERR_ATAPI_SRST_ABORTED     107
#define IDE_ERR_NO_ATAPI_SIG           108
#define IDE_ERR_NOT_ATAPI_DEV          109
#define IDE_ERR_NOT_A_CDROM            110
#define IDE_ERR_WRONG_PKT_SIZE         111
#define IDE_ERR_ATAPI_WRONG_CD_SPEED   112
#define IDE_ERR_COMMAND_ABORTED        113
#define IDE_ERR_UNKNOWN_ERROR          114

//IDE/ATA and ATAPI commands
#define IDE_IDENTIFY_DRIVE             0xec //Also known as IDENTIFY DEVICE
#define IDE_ATAPI_IDENTIFY_DEVICE      0xa1 //ALso known as IDENTIFY PACKET DEVICE
#define IDE_ATAPI_PACKET_COMMAND       0xa0
#define IDE_ATAPI_SOFTRESET            0x08

//ATA register bits
#define IDE_STATUS_ERR                 0x01 //0000 0001
#define IDE_STATUS_BUSY                0x80 //1000 0000
#define IDE_STATUS_DRQ                 0x08 //0000 1000
#define IDE_STATUS_DRDY                0x40 //0100 0000
#define IDE_ERROR_ABRT                 0x04 //0000 0100
#define IDE_DEVCTRL_NIEN               0x02 //0000 0010

//General configuration word (retrieved by IDENTIFY PACKET DEVICE command)
#define IDE_ATAPI_GEN_CFG_WORD_PROTOTYPE       0xc000 //1100 0000 0000 0000, bitmask
#define IDE_ATAPI_GEN_CFG_WORD_DEVTYPE         0x1f00 //0001 1111 0000 0000, bitmask
#define IDE_ATAPI_GEN_CFG_WORD_PKTSIZE         0x0003 //0000 0000 0000 0011, bitmask

#define IDE_ATAPI_GEN_CFG_WORD_PROTOTYPE_ATAPI 0x8000 //1000 0000 0000 0000, 10b = ATAPI device
#define IDE_ATAPI_GEN_CFG_WORD_DEVTYPE_CDROM   0x0500 //0000 0101 0000 0000, 05h = CD-ROM device
#define IDE_ATAPI_GEN_CFG_WORD_PKTSIZE_12B     0x0000 //0000 0000 0000 0000, 00b = 12 bytes

//ATAPI PACKET commands
#define IDE_ATAPI_PKT_SET_CDROM_SPEED  0xbb
#define IDE_ATAPI_PKT_REQUEST_SENSE    0x03

//ATAPI sense-key states
#define IDE_ATAPI_SK_NO_SENSE          0x0
#define IDE_ATAPI_SK_RECOVERED_ERROR   0x1
#define IDE_ATAPI_SK_NOT_READY         0x2
#define IDE_ATAPI_SK_MEDIUM_ERROR      0x3
#define IDE_ATAPI_SK_HARDWARE_ERROR    0x4
#define IDE_ATAPI_SK_ILLEGAL_REQUEST   0x5
#define IDE_ATAPI_SK_UNIT_ATTENTION    0x6
#define IDE_ATAPI_SK_DATA_PROTECT      0x7
#define IDE_ATAPI_SK_ABORTED_COMMAND   0xb
#define IDE_ATAPI_SK_MISCOMPARE        0xe

//ATAPI signature
#define ATAPI_SIG_CYL_HIGH             0xeb
#define ATAPI_SIG_CYL_LOW              0x14

//Devices
#define IDE_DEVICE_MASTER              0x0
#define IDE_DEVICE_SLAVE               0x10

//Memory addresses
#define IDE1_BASE 0x1f0 //Default address for primary controller
#define IDE2_BASE 0x170 //Default address for secondary controller
#define IDE3_BASE 0x1e8 //Default address for tertiary controller
#define IDE4_BASE 0x168 //Default address for quaternary controller

#define IDE_DATA       0
#define IDE_ERROR      1
#define IDE_CYL_HIGH   4
#define IDE_CYL_LOW    5
#define IDE_DRIVE_HEAD 6
#define IDE_STATUS     7
#define IDE_ALT_STATUS 206
#define IDE_DEVCTRL    207


//Prototypes
WORD ide_swap_bytes(WORD w);
int ide_check_if_cdrom(drive_info* di);
int ide_get_atapi_info(PORT port, int device, drive_info* di);
int ide_poll_register(PORT port, BYTE mask, BYTE state);
int ide_send_ata_cmd(PORT port, int device, BYTE opcode, int wait_busy, int wait_ready, int wait_drq);
int ide_check_error_condition(PORT port);
int ide_atapi_packet_set_cd_speed(WORD packet[], int speed);
int ide_atapi_send_packet_cmd(PORT port, int device, WORD packet[]);
int ide_atapi_soft_reset(PORT port, int device);
